package com.qizhen.healsphere.repository.neo4j;

import cn.hutool.core.collection.CollUtil;
import com.qizhen.healsphere.config.Neo4jUtil;
import com.qizhen.healsphere.repository.neo4j.entity.BaseEntity;
import com.qizhen.healsphere.web.dto.NodeDTO;
import org.neo4j.driver.v1.Record;
import org.neo4j.driver.v1.Session;
import org.neo4j.driver.v1.StatementResult;
import org.neo4j.driver.v1.Values;
import org.neo4j.driver.v1.types.Node;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

@Repository
public class BaseNodeRepository {

    @Autowired
    private Neo4jUtil neo4jUtil;

    /**
     * 根据名字和标签查询实体属性
     * @param label
     * @param name
     * @return 实体map
     */
    public Map<String,Object> findNodeByName(String label,String name) {
        String query = String.format("MATCH (n:`%s`) WHERE n.name = $name and (n.is_deleted = 'N' or NOT exists(n.is_deleted)) RETURN n", label);
        try(Session session = neo4jUtil.getSession()) {
            StatementResult result = session.run(query,Values.parameters("name", name));
            Map<String,Object> map = new HashMap<>();
            if(result.hasNext()) {
                Node node = result.single().get("n").asNode();
                map.put("id", node.id());
                map.put("labels", node.labels());
                map.put("properties", node.asMap());
            }
            return map;

        }

    }

    public List<NodeDTO> findNodeByNameLike(String label,String name) {
        String query = String.format("MATCH (n:`%s`) WHERE n.name contains $name and (n.is_deleted = 'N' or NOT exists(n.is_deleted)) return distinct n.name as name,id(n) as id,head(Labels(n)) as label limit 20", label);
        try(Session session = neo4jUtil.getSession()) {
            StatementResult result = session.run(query,Values.parameters("name", name));
            List<NodeDTO> nodeDTOS = new ArrayList<>();
            if(result.hasNext()) {
                List<Record> records = result.list();
                for (Record record : records) {
                    NodeDTO nodeDTO = new NodeDTO();
                    nodeDTO.setName(record.get("name").asString());
                    nodeDTO.setNodeId(record.get("id").asLong());
                    nodeDTO.setLabel(record.get("label").asString());
                    nodeDTOS.add(nodeDTO);
                }
            }
            return nodeDTOS;
        }

    }

    /**
     * 根据id查询实体属性
     * @param id
     * @return 实体map
     */
    public Map<String,Object> findNodeById(Long id) {
        String query = "MATCH (n) WHERE id(n) = $nodeId and (n.is_deleted = 'N' or NOT exists(n.is_deleted)) RETURN n";
        try(Session session = neo4jUtil.getSession()) {
            StatementResult result = session.run(query,Values.parameters("nodeId", id));
            Map<String,Object> map = new HashMap<>();
            if(result.hasNext()) {
                Node node = result.single().get("n").asNode();
                map.put("id", node.id());
                map.put("labels", node.labels());
                map.put("properties", node.asMap());
            }
            return map;

        }

    }

    /**
     * 根据名字模糊查询实体
     * @param label
     * @param name
     * @return 实体map列表
     */
    public List<Map<String,Object>> findNodeListByName(String label, String name) {
        String query = String.format("MATCH (n:`%s`) WHERE n.name contains $name and (n.is_deleted = 'N' or NOT exists(n.is_deleted)) RETURN n", label);

        try(Session session = neo4jUtil.getSession()) {
            StatementResult result = session.run(query,Values.parameters("name", name));
            List<Map<String,Object>> listMap = new ArrayList<>();
            if(result.hasNext()) {
                List<Record> records = result.list();
                for (Record record : records) {
                    Map<String,Object> map = new HashMap<>();
                    Node node = record.get("n").asNode();
                    map.put("id", node.id());
                    map.put("labels", node.labels());
                    map.put("properties", node.asMap());
                    listMap.add(map);
                }
            }
            return listMap;

        }
    }

    /**
     * 新增实体
     * @param label
     * @param properties
     * @return 实体map
     */
    public Map<String, Object> createNode(String label, Map<String, Object> properties) {
        StringBuilder createClause = new StringBuilder();
        if (CollUtil.isEmpty(properties)) {
            properties = new HashMap<>();
        }
        properties.put("is_deleted","N");
        properties.forEach((key, value) -> createClause.append(key).append(": '").append(value).append("'").append(", "));

        // 移除最后的逗号
        if (createClause.length() > 0) {
            createClause.setLength(createClause.length() - 2);
        }

        String query = String.format("MERGE (n:`%s` { %s }) RETURN n", label, createClause);

        try(Session session = neo4jUtil.getSession()) {
            StatementResult result = session.run(query);
            Map<String,Object> map = new HashMap<>();
            if(result.hasNext()) {
                Node node = result.single().get("n").asNode();
                map.put("id", node.id());
                map.put("labels", node.labels());
                map.put("properties", node.asMap());
            }
            return map;
        }
    }

    /**
     * 更新实体名称
     * @param nodeId
     * @param name
     * @return 实体map
     */
    public Map<String, Object> updateNode(Long nodeId, String name) {
        String query = "MATCH (n) WHERE id(n) = $nodeId SET n.name=$name RETURN n";
        try(Session session = neo4jUtil.getSession()) {
            StatementResult result = session.run(query,Values.parameters("nodeId", nodeId,"name", name));
            Map<String,Object> map = new HashMap<>();
            if(result.hasNext()) {
                Node node = result.single().get("n").asNode();
                map.put("id", node.id());
                map.put("labels", node.labels());
                map.put("properties", node.asMap());
            }
            return map;
        }
    }

    /**
     * 删除实体
     * @param nodeId
     */
    public void deleteNode(Long nodeId) {
        String query = "MATCH (n) WHERE id(n) = $nodeId SET n.is_deleted = 'Y' " +
                "WITH n MATCH (n)-[r]-() " +
                "DELETE r";
        try(Session session = neo4jUtil.getSession()) {
            session.run(query,Values.parameters("nodeId", nodeId));
        }
    }

    /**
     * 实体合并
     * @param label
     * @param properties
     * @return 实体map
     */
    public Map<String, Object> mergeNode(String label, Map<String, Object> properties) {
        StringBuilder createClause = new StringBuilder();
        if (CollUtil.isEmpty(properties)) {
            properties = new HashMap<>();
        }
        properties.put("is_deleted","N");
        properties.forEach((key, value) -> createClause.append(key).append(": '").append(value).append("'").append(", "));

        // 移除最后的逗号
        if (createClause.length() > 0) {
            createClause.setLength(createClause.length() - 2);
        }

        String query = String.format("MERGE (n:`%s` { %s }) RETURN n", label, createClause);

        try(Session session = neo4jUtil.getSession()) {
            StatementResult result = session.run(query);
            Map<String,Object> map = new HashMap<>();
            if(result.hasNext()) {
                Node node = result.single().get("n").asNode();
                map.put("id", node.id());
                map.put("labels", node.labels());
                map.put("properties", node.asMap());
            }
            return map;
        }
    }

    /**
     * 新增实体
     * @param label
     * @param properties
     * @return 实体map
     */
    public BaseEntity create(String label, Map<String, Object> properties) {
        StringBuilder createClause = new StringBuilder();
        properties.put("is_deleted","N");
        properties.forEach((key, value) -> createClause.append(key).append(": '").append(value).append("'").append(", "));

        // 移除最后的逗号
        if (createClause.length() > 0) {
            createClause.setLength(createClause.length() - 2);
        }

        String query = String.format("MERGE (n:`%s` { %s }) RETURN n", label, createClause);

        try(Session session = neo4jUtil.getSession()) {
            StatementResult result = session.run(query);

            if(result.hasNext()) {
                return buildBaseEntity(result);
            }
            return null;
        }
    }

    /**
     * 根据名字和标签查询实体属性
     * @param label
     * @param name
     * @return 实体map
     */
    public BaseEntity byName(String label,String name) {
        String query = String.format("MATCH (n:`%s`) WHERE n.name = $name and (n.is_deleted = 'N' or NOT exists(n.is_deleted)) RETURN n", label);
        try(Session session = neo4jUtil.getSession()) {
            StatementResult result = session.run(query,Values.parameters("name", name));
            if(result.hasNext()) {
                return buildBaseEntity(result);
            }
            return null;
        }

    }

    private BaseEntity buildBaseEntity(StatementResult result) {
        BaseEntity baseEntity = new BaseEntity();
        Node node = result.single().get("n").asNode();
        baseEntity.setId(node.id());
        baseEntity.setLabel((ArrayList)node.labels());
        baseEntity.setName(node.get("name").asString());
        return baseEntity;
    }
}